import hashlib
import json
import locale
import shutil
import socket
from collections import deque
from contextlib import contextmanager
from dataclasses import dataclass, replace
from datetime import datetime
from gzip import GzipFile
from http.client import HTTPSConnection, HTTPConnection, HTTPResponse
from pathlib import Path
from socket import AddressFamily
from threading import Lock
from typing import Deque, Dict, Type, Optional, cast, Any, ContextManager, IO, Union, Iterable
from urllib.parse import urlsplit

from pkm.config.configuration import TomlFileConfiguration
from pkm.utils.commons import Closeable, IllegalStateException
from pkm.utils.dicts import get_or_put
from pkm.utils.http.cache_directive import TF_HTTP, CacheDirective
from pkm.utils.http.http_monitors import FetchResourceMonitoredOp, FetchResourceCacheHitEvent, \
    FetchResourceDownloadStartEvent
from pkm.utils.promises import Promise, Deferred
from pkm.utils.properties import clear_cached_properties, cached_property

_next_connection_number = 0
HTTPConnection.debuglevel = 0

# Sad Hacks:
# ------------
# this sad hack is made because http timestamp parsing requires en_us locale, I can probably write my own parser/writer
# to remove this hack
try:
    locale.setlocale(locale.LC_ALL, 'en_US.utf8')
except locale.Error:
    locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')

# this sad hack is made because for some reason,
# many services still has problem with ipv6 (like cloudfront occasionally have)
# and some are just slower to accept connections (like pypi)
original_getaddrinfo = socket.getaddrinfo


# noinspection PyShadowingBuiltins
def get_addr_info_prefer_ipv4(host, port, family=0, type=0, proto=0, flags=0):
    result = original_getaddrinfo(host, port, family, type, proto, flags)
    result.sort(key=lambda it: it[0] == AddressFamily.AF_INET6)
    return result


socket.getaddrinfo = get_addr_info_prefer_ipv4


class HttpException(IOError):

    def __init__(self, msg: str, response: Optional[HTTPResponse] = None) -> None:
        super().__init__(msg)
        self.response = response


@dataclass(frozen=True, eq=True)
class Url:
    scheme: str
    host: str
    port: int
    path: str
    query_string: str
    fragment: str

    def connection_part(self) -> str:
        return f"{self.scheme}://{self.host}:{self.port}"

    def resource_part(self) -> str:
        query_str = "" if not self.query_string else f"?{self.query_string}"
        fragment_str = "" if not self.fragment else f"#{self.fragment}"
        return f"{self.path}{query_str}{fragment_str}"

    def __str__(self):
        default_port = (self.scheme == 'https' and self.port == 443) or (self.scheme == 'http' and self.port == 80)
        port_str = "" if default_port else f":{self.port}"
        query_str = "" if not self.query_string else f"?{self.query_string}"
        fragment_str = "" if not self.fragment else f"#{self.fragment}"
        return f"{self.scheme}://{self.host}{port_str}{self.path}{query_str}{fragment_str}"

    @property
    def connection_type(self) -> Type[HTTPConnection]:
        if self.scheme == 'https':
            return HTTPSConnection
        return HTTPConnection

    @classmethod
    def parse(cls, url: str) -> "Url":
        parts = urlsplit(url)
        scheme = parts.scheme
        if ':' in parts.netloc:
            host, port = parts.netloc.split(':')
        else:
            host = parts.netloc
            port = '80' if scheme == 'http' else '443'

        # if not scheme or not host:
        #     raise ValueError("no schema or host in url")
        return Url(scheme, host, int(port), parts.path, parts.query, parts.fragment)


class _ConnectionProxy(Closeable):
    def __init__(self, conn: HTTPConnection, owner: "_Connections"):
        self._conn = conn
        self._owner = owner

    def __getattr__(self, item):
        if self._conn is None:
            raise ConnectionError("connection closed")

        return getattr(self._conn, item)

    def __del__(self):
        self.close()

    def close(self):
        if self._conn is not None:
            # noinspection PyProtectedMember
            self._owner._release(self._conn)
            self._conn = None


class _Connections(Closeable):
    def __init__(self, ctype: Type[HTTPConnection], host: str, port: int, cached_amount: int):
        self._ctype = ctype
        self._host = host
        self._port = port
        self._connections: Deque[HTTPConnection] = deque()
        self._lock = Lock()
        self._cached_amount = cached_amount
        self._closed = False

    def acquire(self) -> HTTPConnection:
        result: Optional[HTTPConnection] = None
        with self._lock:
            if self._connections:
                result = self._connections.pop()

        if not result:
            result = self._ctype(host=self._host, port=self._port, timeout=10)
            global _next_connection_number
            result.number = _next_connection_number
            _next_connection_number += 1
        return cast(HTTPConnection, _ConnectionProxy(result, self))

    def _release(self, conn: HTTPConnection):
        with self._lock:
            if conn in self._connections:
                return
            if not self._closed and len(self._connections) < self._cached_amount:
                self._connections.append(conn)
            else:
                conn.close()

    def close(self):
        with self._lock:
            self._closed = True
            for c in self._connections:
                c.close()
            self._connections.clear()


class _ConnectionPool(Closeable):

    def __init__(self, cached_amount: int = 24):
        self._connections: Dict[str, _Connections] = {}
        self._cached_amount = cached_amount
        self._lock = Lock()
        self._closed = False

    def connection_for(self, url: Url) -> HTTPConnection:
        connection_key = url.connection_part()
        with self._lock:
            if self._closed:
                raise IllegalStateException("already closed")

            return get_or_put(
                self._connections, connection_key,
                lambda: _Connections(url.connection_type, url.host, url.port, cached_amount=self._cached_amount)) \
                .acquire()

    def close(self):
        with self._lock:
            self._closed = True
            for conn in self._connections.values():
                conn.close()
            self._connections.clear()


@dataclass
class FetchedResource:
    fetch_info: Path
    data: Path
    _root: Path

    def delete(self):
        self.fetch_info.unlink(missing_ok=True)
        self.data.unlink(missing_ok=True)

        p = self.fetch_info.parent
        while p != self._root:
            if any(p.iterdir()):
                break

            p.unlink()
            p = p.parent

        clear_cached_properties(self)

    def exists(self) -> bool:
        return self.data.exists() and self.fetch_info.exists()

    def read_data_as_json(self) -> Any:
        with self.data.open('r') as inp:
            return json.load(inp)

    @cached_property
    def fetch_info_data(self) -> TomlFileConfiguration:
        return TomlFileConfiguration.load(self.fetch_info)

    def is_hash_valid(self, hash_function: str, hash_hex_value: str):

        fetch_info = self.fetch_info_data
        fetch_hash = fetch_info[f'hash.{hash_function}']

        if not fetch_hash:
            hash_ = hashlib.new(hash_function)
            with open(self.data, 'rb') as f:
                while True:
                    next_ = f.read(8192)
                    if not next_:
                        break
                    hash_.update(next_)

            fetch_hash = hash_.hexdigest()
            fetch_info[f'hash.{hash_function}'] = fetch_hash
            fetch_info.save()

        return fetch_hash == hash_hex_value

    def save(self, response: HTTPResponse):
        self.data.parent.mkdir(exist_ok=True, parents=True)
        if self.fetch_info.exists():
            self.fetch_info.unlink()

        try:
            with open(self.data, 'wb') as df:
                shutil.copyfileobj(response, df)

            rheaders = response.headers
            last_modified = rheaders.get('last-modified', rheaders.get('date'))

            # noinspection PyPropertyAccess
            fetch_info = self.fetch_info_data = TomlFileConfiguration(path=self.fetch_info)
            fetch_info['fetch-time'] = last_modified or datetime.utcnow().strftime(TF_HTTP)
            fetch_info['etag'] = rheaders.get("etag") or ''

            fetch_info.save()
        except:  # noqa
            self.delete()
            raise


class _GzipResponseWrapper:
    def __init__(self, response: HTTPResponse):
        self._response = response
        self._stream = GzipFile(fileobj=self._response)

    def read(self, amt: Optional[int] = None):
        return self._stream.read(amt if amt is not None else -1)

    def readinto(self, b: bytearray):
        return self._stream.readinto(b)

    def __getattr__(self, item):
        return getattr(self._response, item)


def _add_standard_headers(headers: Dict[str, str]):
    headers['user-agent'] = f'pkm'
    headers['accept-encoding'] = 'gzip'
    headers['connection'] = 'keep-alive'


class HttpClient:

    def __init__(self, workspace: Path, max_redirects: int = 3, max_connection_retries: int = 2):
        self.workspace = workspace
        workspace.mkdir(exist_ok=True, parents=True)

        self._pool = _ConnectionPool()
        self._fetch_inprogress: Dict[str, Promise[FetchedResource]] = {}
        self._fetch_lock = Lock()
        self._max_redirects = max_redirects
        self._max_connection_retries = max_connection_retries

    def _resource_files_of(self, url: Url) -> FetchedResource:
        url_hash = hashlib.md5(str(url).encode('ascii')).hexdigest()
        url_path = Path(url.path.lstrip('/'))

        if url_path.parent:
            cache_dir = self.workspace / url.host / url_path.parent / url_hash
        else:
            cache_dir = self.workspace / url.host / url_hash

        data_file = cache_dir / url_path.name
        fetch_file = data_file.with_suffix(".fetch.toml")

        return FetchedResource(fetch_file, data_file, self.workspace)

    @contextmanager
    def _request(
            self, mtd: str, url: Url, headers: Dict[str, str],
            payload: Union[IO, str, bytes, None, Iterable[bytes]] = None, *,
            max_redirects: int = -1
    ) -> ContextManager[HTTPResponse]:

        conn: HTTPConnection
        num_redirects_performed = 0
        num_connection_retries = 0

        max_redirects = self._max_redirects if max_redirects < 0 else max_redirects

        while num_connection_retries < self._max_connection_retries and num_redirects_performed <= max_redirects:
            with self._pool.connection_for(url) as conn:
                try:
                    # print(f"[CONN] using connection {conn.number} to connect to {str(url)}")  # noqa
                    conn.request(mtd, url.resource_part(), headers=headers, body=payload)
                    with conn.getresponse() as response:
                        if response.status in (301, 302, 303, 307, 308):
                            num_redirects_performed += 1
                            num_connection_retries = 0
                            if not (new_location := response.headers.get("Location")):
                                raise HttpException(
                                    "server responded with redirect but without supplying a new location")
                            if new_location.startswith("/"):
                                url = replace(url, path=new_location)
                            else:
                                url = Url.parse(new_location)
                            response.read()
                        else:
                            if response.headers.get('Content-Encoding') == 'gzip':
                                yield cast(HTTPResponse, _GzipResponseWrapper(response))
                            else:
                                yield response
                            return
                except ConnectionError:
                    num_connection_retries += 1

        error = "max redirects reached" if num_redirects_performed > self._max_redirects else "connection failed"
        raise HttpException(error)

    @contextmanager
    def post(
            self, url: str, payload: Union[IO, str, bytes, None, Iterable[bytes]] = None,
            headers: Optional[Dict[str, str]] = None, *, max_redirects: int = -1) -> ContextManager[HTTPResponse]:

        headers = headers or {}
        url = Url.parse(url)
        _add_standard_headers(headers)

        with self._request("POST", url, headers, payload, max_redirects=max_redirects) as response:
            yield response

    @contextmanager
    def get(self, url: str, headers: Optional[Dict[str, str]] = None, *,
            max_redirects: int = -1) -> ContextManager[HTTPResponse]:

        headers = headers or {}
        url = Url.parse(url)
        _add_standard_headers(headers)

        with self._request("GET", url, headers, max_redirects=max_redirects) as response:
            yield response

    def fetch_resource(
            self, url: str, cache: Optional[CacheDirective] = None,
            resource_name: Optional[str] = None) -> Optional[FetchedResource]:

        parsed_url = Url.parse(url)
        resource_name = resource_name or parsed_url.path.split("/")[-1]

        with FetchResourceMonitoredOp(resource_name, url) as mop:
            if not cache:
                cache = CacheDirective.allways()

            cached = True
            with self._fetch_lock:

                # If I am already requesting this url - no need to actually do it twice, note that
                # this may pose a problem if one of the recipients will delete the files, if it revealed to be a problem
                # we can fix it by adding a ref-counting delete operation for the FetchedResource for example
                promise = self._fetch_inprogress.get(url)

                if not promise:
                    cached = False

                    def _fetch() -> Optional[FetchedResource]:
                        headers = {}
                        cache_files = self._resource_files_of(parsed_url)
                        fetch_info = TomlFileConfiguration.load(
                            cache_files.fetch_info)  # TODO: maybe add the response headers
                        _add_standard_headers(headers)

                        use_cached_data = cache_files.exists() and cache.is_cache_valid(fetch_info)
                        if use_cached_data:
                            mop.notify(FetchResourceCacheHitEvent())
                        else:
                            # console.log(f"[Start]: GET {url}...")
                            response: Optional[HTTPResponse] = None

                            # noinspection PyShadowingNames
                            try:
                                cache.add_headers(fetch_info, headers)
                                with self._request("GET", parsed_url, headers=headers) as response:
                                    clength = int(response.headers.get('content-length', '-1'))
                                    mop.notify(FetchResourceDownloadStartEvent(clength, cache_files.data))

                                    # print(f"[MID] GET {url} resulted with status code: {response.status}...")

                                    if response.status == 200:
                                        cache_files.save(response)
                                    elif response.status != 304:
                                        raise HttpException(
                                            f"request to {url} ended with unexpected status code:"
                                            f" {response.status} ({response.msg})", response)
                            except Exception as e:
                                if isinstance(e, HttpException):
                                    raise
                                else:
                                    raise HttpException(str(e), response) from e
                            finally:
                                # print(f"thread: {threading.current_thread().ident} attempting to release task")
                                with self._fetch_lock:
                                    del self._fetch_inprogress[url]
                                # print(f"thread: {threading.current_thread().ident} succss to release task")

                        # print(f"[END] GET {url} resulted with cache files: {cache_files}")
                        return cache_files

                    deffered: Deferred[Optional[FetchedResource]] = Deferred()
                    promise = self._fetch_inprogress[url] = deffered.promise()

            if cached:
                mop.notify(FetchResourceCacheHitEvent())
                return promise.result()

            try:
                deffered.complete(_fetch())
            except Exception as e:
                deffered.fail(e)

            return promise.result()

    def clear_resources(self):
        shutil.rmtree(self.workspace)
        self.workspace.mkdir(parents=True, exist_ok=True)



from pathlib import Path
from typing import List, Optional, Iterable

from pkm.api.dependencies.dependency import Dependency
from pkm.api.distributions.distribution import Distribution
from pkm.api.environments.environment import Environment
from pkm.api.packages.package import Package
from pkm.api.projects.project import Project
from pkm.api.repositories.repository import AbstractRepository
from pkm.utils.strings import endswith_any, without_prefix


class FileRepository(AbstractRepository):
    def __init__(self):
        super().__init__("file")

    def _do_match(self, dependency: Dependency, env: Environment) -> List[Package]:
        if not (vurl := dependency.required_url()):
            return []

        package = load_package_from_filesystem(Path(without_prefix(vurl.url, "file://")))
        return [package] if package else []

    def accepted_url_protocols(self) -> Iterable[str]:
        return 'file',


def load_package_from_filesystem(location: Path) -> Optional[Package]:
    if location.is_dir():
        return Project.load(location)
    elif endswith_any(location.name, ('.tar.gz', '.whl')):
        return Distribution.package_from(location)

    return None



from __future__ import annotations
from copy import copy
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Any, Dict, List

from pkm.config.configuration import TomlFileConfiguration, computed_based_on
from pkm.utils.dicts import remove_none_values
from pkm.utils.hashes import HashBuilder


class RepositoriesConfigInheritanceMode(Enum):
    INHERIT_CONTEXT = 0,
    INHERIT_MAIN = 1,
    NO_INHERITANCE = 2


class RepositoriesConfiguration(TomlFileConfiguration):
    repositories: List[RepositoryInstanceConfig]
    inheritance_mode: RepositoriesConfigInheritanceMode

    @computed_based_on("inheritance")
    def inheritance_mode(self) -> RepositoriesConfigInheritanceMode:
        value = self["inheritance"]
        if value is not None:
            if value == "main":
                return RepositoriesConfigInheritanceMode.INHERIT_MAIN
            if not value:
                return RepositoriesConfigInheritanceMode.NO_INHERITANCE
        return RepositoriesConfigInheritanceMode.INHERIT_CONTEXT

    @computed_based_on("")
    def repositories(self) -> List[RepositoryInstanceConfig]:
        return [
            RepositoryInstanceConfig.from_config(name, repo) for name, repo in self.items()
            if isinstance(repo, Dict)]

    @repositories.modifier
    def set_repositories(self, repos: List[RepositoryInstanceConfig]):
        for k, v in list(self.items()):
            if isinstance(v, Dict):
                del self[k]
        for repo in repos:
            self[repo.name] = repo.to_config()


@dataclass(frozen=True, eq=True)
class RepositoryInstanceConfig:
    type: str
    packages: Optional[List[str]]  # use to limit to the repository to only the given packages
    name: Optional[str]
    args: Dict[str, str]

    def __hash__(self):
        return HashBuilder() \
            .regulars(self.type, self.name) \
            .unordered_mapping(self.args) \
            .unordered_seq(self.packages) \
            .build()

    def to_config(self) -> Dict[str, Any]:
        return remove_none_values({
            **self.args,
            'type': self.type,
            'packages': self.packages,
        })

    @classmethod
    def from_config(cls, name: str, config: Dict[str, Any]) -> "RepositoryInstanceConfig":
        config = copy(config)
        type_ = config.pop('type')
        packages: Optional[List[str]] = config.pop('packages', None)
        args = config
        return RepositoryInstanceConfig(type_, packages, name, args)



from typing import Iterable, Optional


def without_suffix(s: str, suffix: str) -> str:
    """
    :param s: the base string
    :param suffix: the suffix to remove
    :return: the value of `s` without `suffix` if `s` ends with `suffix` otherwise unchanged `s`
    """
    if s.endswith(suffix):
        return s[:-len(suffix)]

    return s


def without_prefix(s: str, prefix: str) -> str:
    """
    :param s: the base string
    :param prefix: the prefix to remove
    :return: the value of `s` without `prefix` if `s` starts with `prefix` otherwise unchanged `s`
    """
    if s.startswith(prefix):
        return s[len(prefix):]

    return s


def endswith_any(s: str, suffixes: Iterable[str]) -> Optional[str]:
    """
    :param s: the string to check
    :param suffixes: iterable of suffixes to search
    :return: the first suffix of `s` in `suffixes` or `None` if no such suffix found
    """
    return next((sf for sf in suffixes if s.endswith(sf)), None)


def startswith_any(s: str, prefixes: Iterable[str]) -> Optional[str]:
    """
    :param s: the string to check
    :param prefixes: iterable of prefixes to search
    :return: the first prefix of `s` in `prefixes` or `None` if no such prefix found
    """
    return next((pf for pf in prefixes if s.startswith(pf)), None)



from abc import abstractmethod
from dataclasses import dataclass, field
from io import UnsupportedOperation
from pathlib import Path
from typing import Optional, Any, Dict, List, TYPE_CHECKING

from pkm.api.dependencies.dependency import Dependency
from pkm.api.distributions.distinfo import PackageInstallationInfo
from pkm.api.distributions.source_distribution import SourceDistribution
from pkm.api.distributions.wheel_distribution import WheelDistribution
from pkm.api.packages.package import Package, PackageDescriptor
from pkm.api.packages.package_installation import PackageInstallationTarget
from pkm.api.packages.package_metadata import PackageMetadata
from pkm.api.packages.package_monitors import PackageInstallMonitoredOp
from pkm.api.versions.version_specifiers import VersionSpecifier
from pkm.utils.hashes import HashSignature
from pkm.utils.types import SupportsLessThanEq

if TYPE_CHECKING:
    from pkm.api.environments.environment import Environment


@dataclass(frozen=True, eq=True)
class PackageArtifact:
    file_name: str
    requires_python: Optional[VersionSpecifier] = None
    other_info: Dict[str, Any] = field(default_factory=dict)
    hash_sig: Optional[HashSignature] = None

    def is_binary(self):
        return self.file_name.endswith(".whl")


class AbstractPackage(Package):

    def __init__(self, descriptor: PackageDescriptor, artifacts: List[PackageArtifact],
                 published_metadata: Optional[PackageMetadata] = None):
        self._descriptor = descriptor
        self._artifacts = artifacts
        self._dependencies_per_artifact_id: Dict[int, List["Dependency"]] = {}
        self._path_per_artifact_id: Dict[int, Path] = {}
        self._published_metadata = published_metadata

    @property
    def published_metadata(self) -> Optional[PackageMetadata]:
        return self._published_metadata

    @property
    def descriptor(self) -> PackageDescriptor:
        return self._descriptor

    def best_artifact_for(self, env: "Environment") -> Optional[PackageArtifact]:
        """
        searches for the best artifact that compatible with `env`
        :param env: the environment that the searched artifact should be compatible with
        :return: the best artifact found or None if no compatible artifact exists
        """

        env_interpreter = env.interpreter_version

        source_dist: Optional[PackageArtifact] = None
        best_binary_dist: Optional[PackageArtifact] = None
        best_binary_dist_score: Optional[SupportsLessThanEq] = None

        for artifact in self._artifacts:
            requires_python = artifact.requires_python

            file_name: str = artifact.file_name
            is_binary = artifact.is_binary()

            if requires_python:
                try:
                    if not requires_python.allows_version(env_interpreter):
                        continue
                except ValueError:
                    print(
                        'could not parse python requirements for artifact:'
                        f' {file_name} of {self.name}=={self.version}, skipping it')
                    continue

            if is_binary:
                ctag = WheelDistribution.extract_compatibility_tags_of(Path(file_name))
                if (score := env.compatibility_tag_score(ctag)) is not None:
                    if best_binary_dist_score is None or best_binary_dist_score < score:
                        best_binary_dist = artifact
                        best_binary_dist_score = score

            else:
                source_dist = artifact

        return best_binary_dist or source_dist

    def is_compatible_with(self, env: "Environment"):
        return self.best_artifact_for(env) is not None

    @abstractmethod
    def _retrieve_artifact(self, artifact: PackageArtifact) -> Path:
        """
        retrieve the given artifact, storing it to the file system and returning it
        :param artifact: the artifact to retrieve
        :return: the stored artifact
        """

    def install_to(
            self, target: PackageInstallationTarget, user_request: Optional["Dependency"] = None,
            editable: bool = False):

        with PackageInstallMonitoredOp(self.descriptor):
            artifact = self.best_artifact_for(target.env)
            artifact_path = self._get_or_retrieve_artifact_path(artifact)
            if hashsig := artifact.hash_sig:
                if not hashsig.validate_against(artifact_path):
                    raise ValueError(f"Security Risk: invalid hash for {self.descriptor}")

            installation_mode = PackageInstallationInfo(False, editable)
            if artifact.is_binary():
                WheelDistribution(self.descriptor, artifact_path).install_to(target, user_request, installation_mode)
            else:
                SourceDistribution(self.descriptor, artifact_path).install_to(target, user_request, installation_mode)

    def _get_or_retrieve_artifact_path(self, artifact: PackageArtifact):
        if not (artifact_path := self._path_per_artifact_id.get(id(artifact))):
            artifact_path = self._retrieve_artifact(artifact)

            self._path_per_artifact_id[id(artifact)] = artifact_path
        return artifact_path

    def dependencies(
            self, environment: "Environment", extras: Optional[List[str]] = None) -> List["Dependency"]:
        all_deps = self._unfiltered_dependencies(environment)
        return [d for d in all_deps if d.is_applicable_for(environment, extras)]

    def _unfiltered_dependencies(self, environment: "Environment") -> List["Dependency"]:
        artifact = self.best_artifact_for(environment)
        if not artifact:
            raise UnsupportedOperation(
                "attempting to compute dependencies for environment that is not supported by this package")

        if deps := self._dependencies_per_artifact_id.get(id(artifact)):
            return deps
        resource = self._get_or_retrieve_artifact_path(artifact)
        filename = artifact.file_name
        if filename.endswith('.whl'):
            info = WheelDistribution(self.descriptor, resource) \
                .extract_metadata(environment)
        else:
            info = SourceDistribution(self.descriptor, resource) \
                .extract_metadata(environment)

        self._dependencies_per_artifact_id[id(artifact)] = info.dependencies
        return info.dependencies



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import List, Dict, Any, Optional, Iterable

from pkm.api.dependencies.dependency import Dependency
from pkm.api.environments.environment import Environment
from pkm.api.packages.package import PackageDescriptor, Package
from pkm.api.repositories.repository import AbstractRepository, Repository
from pkm.config.configuration import TomlFileConfiguration
from pkm.utils.commons import unone
from pkm.utils.iterators import groupby


@dataclass
class _LockedVersion:
    env_markers_hash: str
    package: PackageDescriptor

    def write(self) -> Dict[str, Any]:
        return {
            'env_markers_hash': self.env_markers_hash,
            'package': self.package.write(),
        }

    @classmethod
    def read(cls, data: Dict[str, Any]) -> "_LockedVersion":
        return _LockedVersion(data['env_markers_hash'], PackageDescriptor.read(data['package']))


class PackagesLock:
    """
    represent a multi-environment packages lock
    this lock type can handle projects that are being worked on by different users, each using different environment
    (e.g., different os, interpreter version, etc.) for the development process without them breaking each other lock
    repeatedly
    """

    def __init__(self, locked_packages: Optional[List[_LockedVersion]] = None, lock_file: Optional[Path] = None):
        locked_packages = unone(locked_packages, list)
        self._locked_packages: Dict[str, List[_LockedVersion]] = \
            groupby(locked_packages, lambda it: it.package.name)
        self._lock_file = lock_file

    def unlock_packages(self, package_names: Iterable[str]) -> PackagesLock:
        """
        unlock the given package names, removing their information from the lock
        (does not automatically save the lock after this operation)
        :param package_names: the packages to unlock
        :return self (for chaining support)
        """
        for pn in package_names:
            self._locked_packages.pop(pn, None)

        return self

    def env_specific_locks(self, env: Environment):
        """
        :param env: the environment to extract lock information to
        :return: list of all locked packages that are specific to the given environment
        """
        env_markers_hash = env.markers_hash
        return [lp.package for lst in self._locked_packages.values() for lp in lst if
                lp.env_markers_hash == env_markers_hash]

    def locked_versions(self, env: Environment, package: str) -> List[PackageDescriptor]:
        """
        :param env: environment that the given dependency should be compatible with
        :param package: the package to find lock information for
        :return: list of package descriptors to try and install on the environment, sorted by importance
                 (try the first one first)
        """

        relevant_locks = [lp for lst in self._locked_packages.values() for lp in lst] if not package else \
            [lp for lp in (self._locked_packages.get(package) or ())]

        if not relevant_locks:
            return []

        env_markers_hash = env.markers_hash

        # the following lines may look wrong but what we are actually trying to achieve here is
        # to try and install the env-specific locked package first if such exists otherwise we would like to
        # try and install non-env specific pre-locked packages
        result = [lock.package for lock in relevant_locks if lock.env_markers_hash != env_markers_hash]
        if len(result) != len(relevant_locks):
            result.insert(0, next(lock.package for lock in relevant_locks if lock.env_markers_hash == env_markers_hash))

        return result

    def update_lock(self, env: Environment):
        """
        lock the packages in the given environment
        :param env: the environment to use to extract lock information
        """

        env_hash = env.markers_hash
        new_locks = [
            lock for locks_by_name in self._locked_packages.values()
            for lock in locks_by_name
            if lock.env_markers_hash != env_hash]

        for package in env.site_packages.installed_packages():
            new_locks.append(_LockedVersion(env_hash, package.descriptor))

        self._locked_packages = groupby(new_locks, lambda it: it.package.name)

    def save(self, lock_file: Optional[Path] = None):
        """
        saves the state of this packages lock to the given file
        :param lock_file:
        """

        lock_file = lock_file or self._lock_file
        if not lock_file:
            raise FileNotFoundError('lock file is not given')

        configuration = TomlFileConfiguration.load(lock_file)
        locks = [lp.write() for locks_by_name in self._locked_packages.values() for lp in locks_by_name]
        configuration['lock'] = locks
        configuration.save()

    def sort_packages_by_lock_preference(self, env: Environment, packages: List[Package]) -> List[Package]:
        if packages:
            locked_versions = {lp.version for lp in self.locked_versions(env, packages[0].name)}
            packages.sort(key=lambda it: 0 if it.version in locked_versions else 1)

        return packages

    @classmethod
    def load(cls, lock_file: Path) -> "PackagesLock":
        """
        load packages lock from the given lock file
        :param lock_file: the file to load from
        :return: the loaded lock
        """

        configuration = TomlFileConfiguration.load(lock_file)
        locked_packages = [_LockedVersion.read(lp) for lp in (configuration['lock'] or [])]
        return PackagesLock(locked_packages, lock_file)


class LockPrioritizingRepository(AbstractRepository):
    def __init__(self, name: str, base_repo: Repository, lock: PackagesLock, env: Environment):
        super().__init__(name)
        self._base_repo = base_repo
        self._lock = lock
        self._env = env

    def _do_match(self, dependency: Dependency, env: Environment) -> List[Package]:
        return self._base_repo.match(dependency, env)

    def _sort_by_priority(self, dependency: Dependency, packages: List[Package]) -> List[Package]:
        return self._lock.sort_packages_by_lock_preference(self._env, packages)

    def accepted_url_protocols(self) -> Iterable[str]:
        return self._base_repo.accepted_url_protocols()

    def accept_non_url_packages(self) -> bool:
        return self._base_repo.accept_non_url_packages()



from __future__ import annotations
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Iterable

from pkm.api.dependencies.dependency import Dependency
from pkm.api.environments.environment import Environment
from pkm.api.environments.environments_zoo import EnvironmentsZoo
from pkm.api.packages.package import PackageDescriptor, Package
from pkm.api.projects.project import Project
from pkm.api.projects.project_group import ProjectGroup
from pkm.api.repositories.repositories_configuration import RepositoriesConfiguration, RepositoryInstanceConfig
from pkm.api.repositories.repository import Repository, AbstractRepository
from pkm.api.repositories.repository_loader import RepositoryLoader, REPOSITORIES_CONFIGURATION_PATH
from pkm.repositories.shared_pacakges_repo import SharedPackagesRepository
from pkm.resolution.packages_lock import LockPrioritizingRepository
from pkm.utils.properties import cached_property, clear_cached_properties


class RepositoryManagement(ABC):
    def __init__(self, cfg: RepositoriesConfiguration, loader: Optional[RepositoryLoader] = None):
        if not loader:
            from pkm.api.pkm import pkm
            loader = pkm.repository_loader

        self._loader = loader
        self.configuration = cfg

    @abstractmethod
    def _load_attached(self) -> Repository:
        ...

    @cached_property
    def attached_repo(self) -> Repository:
        return self._load_attached()

    def _update_config(self):
        self.configuration.save()
        clear_cached_properties(self)

    def add_repository(self, name: str, builder: str, args: Dict[str, str], packages_limit: Optional[List[str]] = None):
        config = RepositoryInstanceConfig(builder, packages_limit, name, args)

        # the following line act a s a safeguard before the add operation,
        # if it fails the actual addition will not take place
        self._loader.build(config)

        self.configuration.repositories = [
            *self.configuration.repositories, config]

        self._update_config()

    def list_defined_repositories(self) -> List[RepositoryInstanceConfig]:
        return self.configuration.repositories

    def remove_repository(self, name: str):
        self.configuration.repositories = [repo for repo in self.configuration.repositories if repo.name != name]
        self._update_config()


def _config_at(path: Path) -> RepositoriesConfiguration:
    return RepositoriesConfiguration.load(path / REPOSITORIES_CONFIGURATION_PATH)


class EnvRepositoryManagement(RepositoryManagement):

    def __init__(self, env: Environment, loader: Optional[RepositoryLoader] = None):
        super().__init__(_config_at(env.path), loader)
        self.env = env

    def _load_attached(self) -> Repository:
        repo = self._loader.main
        if zoo := self.env.zoo:
            repo = zoo.attached_repository

        if self.configuration.path.exists():
            repo = self._loader.load("env-configured-repository", self.configuration, repo)

        return repo


class ZooRepositoryManagement(RepositoryManagement):
    def __init__(self, zoo: EnvironmentsZoo, loader: Optional[RepositoryLoader] = None):
        super().__init__(_config_at(zoo.path), loader)
        self.zoo = zoo

    def _load_attached(self) -> Repository:
        repo = self._loader.main
        if self.configuration.path.exists():
            repo = self._loader.load("zoo-configured-repository", self.configuration, repo)

        if self.zoo.config.package_sharing.enabled:
            repo = SharedPackagesRepository(self.zoo.path / ".zoo/shared", repo)

        return repo


def _load_for_project_group(
        loader: RepositoryLoader,
        group: ProjectGroup,
        config: RepositoriesConfiguration,
        base_repo: Optional[Repository] = None) -> Repository:
    repo = base_repo or loader.main

    if config.path.exists():
        repo = loader.load("group-configured-repository", config, repo)

    repo = _ProjectsRepository.create("group-projects-repository", group.project_children_recursive, repo)
    return repo


class ProjectRepositoryManagement(RepositoryManagement):

    def __init__(self, project: Project, loader: Optional[RepositoryLoader] = None):
        super().__init__(_config_at(project.path), loader)
        self.project = project

    def _load_attached(self) -> Repository:
        repo = self.project.attached_environment.attached_repository

        if group := self.project.group:
            repo = _load_for_project_group(self._loader, group, _config_at(group.path), repo)
        else:
            repo = _ProjectsRepository.create('project-repository', [self.project], repo)

        if self.configuration.exists():
            repo = self._loader.load("project-configured-repository", self.configuration, repo)

        repo = LockPrioritizingRepository(
            "lock-prioritizing-repository", repo, self.project.lock,
            self.project.attached_environment)

        return repo


class ProjectGroupRepositoryManagement(RepositoryManagement):
    def __init__(self, group: ProjectGroup, loader: Optional[RepositoryLoader] = None):
        super().__init__(_config_at(group.path), loader)
        self.group = group

    def _load_attached(self) -> Repository:
        return _load_for_project_group(self._loader, self.group, self.configuration, None)


class _ProjectsRepository(AbstractRepository):
    def __init__(self, name: str, projects: Dict[str, Tuple[PackageDescriptor, Path]], base_repo: Repository):
        super().__init__(name)
        self._packages = projects
        self._base_repo = base_repo

    def _do_match(self, dependency: Dependency, env: Environment) -> List[Package]:
        if (package_and_path := self._packages.get(dependency.package_name)) and \
                dependency.version_spec.allows_version(package_and_path[0].version):
            return [Project.load(package_and_path[1])]
        return self._base_repo.match(dependency, env)

    def _sort_by_priority(self, dependency: Dependency, packages: List[Package]) -> List[Package]:
        return packages

    @classmethod
    def create(cls, name: str, projects: Iterable[Project], base: Repository) -> _ProjectsRepository:
        return _ProjectsRepository(name, {p.name: (p.descriptor, p.path) for p in projects}, base)

    def accept_non_url_packages(self) -> bool:
        return self._base_repo.accept_non_url_packages()

    def accepted_url_protocols(self) -> Iterable[str]:
        return self._base_repo.accepted_url_protocols()



from __future__ import annotations

import re
from dataclasses import dataclass
from functools import reduce
from importlib import import_module
from types import ModuleType
from typing import ClassVar, Optional, List, Any

from pkm.utils.commons import UnsupportedOperationException

# noinspection RegExpRedundantEscape
_OBJ_REF_RX = re.compile(r"(?P<mdl>[^:]*)(:(?P<obj>[^\[]*)\s*(\[\s*(?P<ext>[^\]]*)\s*\])?)?")
_EXT_DELIM_RX = re.compile("\\s*,\\s*")


@dataclass(frozen=True)
class ObjectReference:
    """
    The object reference points to a Python object. It is either in the form importable.module,
    or importable.module:object.attr.
    Each of the parts delimited by dots and the colon is a valid Python identifier.

    read more about it here: https://packaging.python.org/en/latest/specifications/entry-points/
    """
    module_path: str
    object_path: Optional[str] = None
    extras: Optional[List[str]] = None

    def __str__(self):
        obj_str = f":{self.object_path}" if self.object_path else ""
        ext_str = f" [{', '.join(self.extras)}]" if self.extras else ""
        return f"{self.module_path}{obj_str}{ext_str}"

    def import_module(self) -> ModuleType:
        return import_module(self.module_path)

    def import_object(self) -> Any:
        module = self.import_module()
        return reduce(getattr, self.object_path.split("."), module)

    def execution_script_snippet(self) -> str:
        if not self.object_path:
            raise UnsupportedOperationException(
                f"{str(self)} cannot be used as script generator - "
                f"it must be of the format module.path:zero.arg.function.path")

        return f"import sys;import {self.module_path};sys.exit({self.module_path}.{self.object_path}())"

    @classmethod
    def parse(cls, refstr: str) -> "ObjectReference":
        if not (match := _OBJ_REF_RX.match(refstr)):
            raise ValueError(f"could not parse object reference from string: '{refstr}'")

        module_path = match.group('mdl')
        object_path = match.group('obj')
        extras = _EXT_DELIM_RX.split(ext.strip()) if (ext := match.group('ext')) else None

        return ObjectReference(module_path, object_path, extras)


@dataclass
class EntryPoint:
    """
    Entry points are a mechanism for an installed distribution to advertise components it provides to be discovered and
    used by other code.
    read more about it here: https://packaging.python.org/en/latest/specifications/entry-points/
    """

    group: str
    name: str
    ref: "ObjectReference"

    G_CONSOLE_SCRIPTS: ClassVar[str] = "console_scripts"
    G_GUI_SCRIPTS: ClassVar[str] = "gui_scripts"

    def is_script(self) -> bool:
        """
        :return: true if this entrypoint belongs to one of the script groups, false otherwise
        """

        return self.group in (EntryPoint.G_CONSOLE_SCRIPTS, EntryPoint.G_GUI_SCRIPTS)



from email.message import EmailMessage
from email.parser import Parser
from pathlib import Path
from typing import List, TYPE_CHECKING, Optional

from pkm.api.dependencies.dependency import Dependency

from pkm.api.versions.version import StandardVersion, Version
from pkm.api.versions.version_specifiers import VersionSpecifier, AllowAllVersions
from pkm.config.configuration import FileConfiguration, computed_based_on
from pkm.utils.dicts import get_or_put, remove_none_values

if TYPE_CHECKING:
    from pkm.api.projects.pyproject_configuration import ProjectConfig

_METADATA_VERSION = '2.1'

_MULTI_FIELDS = {
    'Provides-Extra', 'Platform', 'Supported-Platform', 'Classifier', 'Requires-Dist', 'Provides-Dist',
    'Obsoletes-Dist', 'Requires-External', 'Project-URL', 'Dynamic'
}


class PackageMetadata(FileConfiguration):
    dependencies: List[Dependency]
    required_python_spec: VersionSpecifier

    """
    implementation of the package metadata 2.1 specification as described in pep 566
    see also: https://packaging.python.org/en/latest/specifications/core-metadata/
    """

    @computed_based_on("Metadata-Version")
    def metadata_version(self) -> StandardVersion:
        return StandardVersion.parse(self['Metadata-Version'])

    @property
    def package_name(self) -> str:
        return self['Name']

    @property
    def summary(self) -> Optional[str]:
        return self['Summary']

    @property
    def description(self) -> Optional[str]:
        return self['Description']

    @computed_based_on("Version")
    def package_version(self) -> Version:
        return Version.parse(self['Version'])

    @computed_based_on("Requires-Dist")
    def dependencies(self) -> List[Dependency]:
        requires_dist = self["Requires-Dist"] or []
        return [Dependency.parse(d) for d in requires_dist]

    @computed_based_on("Requires-Python")
    def required_python_spec(self) -> VersionSpecifier:
        if version_str := self["Requires-Python"]:
            return VersionSpecifier.parse(version_str)
        return AllowAllVersions

    def generate_content(self) -> str:
        msg = EmailMessage()
        for key, value in self._data.items():
            if value is None:
                continue

            if key in _MULTI_FIELDS and not isinstance(value, List):
                raise ValueError(f'{key} expected to be a list, found: {value}')

            if key == 'Description':
                continue  # will add it in payload

            if isinstance(value, List):
                for item in value:
                    msg[key] = item
            else:
                msg[key] = value

        if payload := self._data.get("Description"):
            msg.set_payload(payload)

        return msg.as_string()

    @classmethod
    def from_project_config(cls, prjc: "ProjectConfig") -> "PackageMetadata":

        # filling authors and maintainers according to pep-621:
        # 1. If only name is provided, the value goes in Author/Maintainer as appropriate.
        # 2. If only email is provided, the value goes in Author-email/Maintainer-email as appropriate.
        # 3. If both email and name are provided, the value goes in Author-email/Maintainer-email as appropriate,
        #    with the format {name} <{email}> (with appropriate quoting, e.g. using email.headerregistry.Address).
        # 4. Multiple values should be separated by commas.

        author_names = [a.name for a in (prjc.authors or []) if a.name]
        author_emails = [a.email for a in (prjc.authors or []) if a.email]

        maintainer_names = [a.name for a in (prjc.authors or []) if a.name]
        maintainer_emails = [a.email for a in (prjc.authors or []) if a.email]

        extras = list((prjc.optional_dependencies or {}).keys())
        all_deps = prjc.all_dependencies

        data = {
            'Metadata-Version': _METADATA_VERSION,
            'Name': prjc.name, 'Version': str(prjc.version),
            'Summary': prjc.description, 'Description': prjc.readme_content(),
            'Description-Content-Type': prjc.readme_content_type(), 'Keywords': prjc.keywords or [],
            'Author': ', '.join(author_names), 'Author-email': ', '.join(author_emails),
            'Maintainer': ', '.join(maintainer_names), 'Maintainer-email': ', '.join(maintainer_emails),
            'License': prjc.license_content(), 'Classifier': prjc.classifiers,
            'Requires-Dist': [str(d) for d in all_deps], 'Requires-Python': str(prjc.requires_python),
            'Provides-Extra': extras,
            # 'Dynamic': prjc.dynamic or [], add back when pypi supports metadata version 2.2
        }

        return PackageMetadata(data=remove_none_values(data), path=None)

    @classmethod
    def load(cls, path: Path) -> "PackageMetadata":
        if not path.exists():
            return PackageMetadata(path=path, data={})

        parser = Parser()
        with path.open('r', errors="ignore") as path_fd:
            content = parser.parse(path_fd)

            data = {}
            for key, value in content.items():
                if key in _MULTI_FIELDS:
                    get_or_put(data, key, list).append(value)
                elif old_value := data.get(key):
                    if isinstance(old_value, List):
                        old_value.append(value)
                    else:
                        data[key] = [old_value, value]
                else:
                    data[key] = value

        return PackageMetadata(path=path, data=data)



from dataclasses import FrozenInstanceError
from threading import RLock

from typing import TypeVar, Callable, Any, Dict

_T = TypeVar("_T")


# noinspection PyPep8Naming
class cached_property:

    def __init__(self, func: Callable[[Any], _T]):
        self._func = func
        self._attr = None
        self.__doc__ = func.__doc__
        self._mutation_lock = RLock()
        self._instance_locks: Dict[int, RLock] = {}

    def __set_name__(self, owner, name):
        self._attr = f"_cached_{name}"

    def _compute(self, instance):
        compute = True
        iid = id(instance)
        with self._mutation_lock:
            instance_lock = self._instance_locks.get(iid)
            if instance_lock is not None:
                compute = False
            else:
                instance_lock = self._instance_locks[iid] = RLock()
                instance_lock.acquire()

        with instance_lock:
            if compute:
                try:
                    if not hasattr(instance, self._attr):
                        value = self._func(instance)
                        try:
                            setattr(instance, self._attr, value)
                        except FrozenInstanceError:
                            super(instance.__class__, instance).__setattr__(self._attr, value)
                finally:
                    del self._instance_locks[iid]
                    instance_lock.release()

    def __get__(self, instance, owner) -> _T:
        if instance is None:
            return self

        while True:
            try:
                return getattr(instance, self._attr)
            except AttributeError:
                self._compute(instance)

    def __set__(self, instance, value: _T):
        setattr(instance, self._attr, value)

    def __delete__(self, instance):
        if hasattr(instance, self._attr):
            delattr(instance, self._attr)


def clear_cached_properties(obj: Any):
    for attr in dir(obj.__class__):
        if isinstance(getattr(obj.__class__, attr), cached_property):
            delattr(obj, attr)
